@@ -38,10 +38,12 @@ function Module(id, parent) {
   this.id = id;
   this.exports = {};
   this.parent = parent;
+  if (parent && parent.children) {
+    parent.children.push(this);
+  }
 
   this.filename = null;
   this.loaded = false;
-  this.exited = false;
   this.children = [];
 }
 module.exports = Module;
@@ -156,7 +158,6 @@ function tryExtensions(p, exts) {
 
 
 Module._findPath = function(request, paths) {
-  var fs = NativeModule.require('fs');
   var exts = Object.keys(Module._extensions);
 
   if (request.charAt(0) === '/') {
@@ -334,7 +335,9 @@ Module._resolveFilename = function(request, parent) {
 
   var filename = Module._findPath(request, paths);
   if (!filename) {
-    throw new Error("Cannot find module '" + request + "'");
+    var err = new Error("Cannot find module '" + request + "'");
+    err.code = 'MODULE_NOT_FOUND';
+    throw err;
   }
   return filename;
 };
@@ -356,6 +359,8 @@ Module.prototype.load = function(filename) {
 
 
 Module.prototype.require = function(path) {
+  assert(typeof path === 'string', 'path must be a string');
+  assert(path, 'missing path');
   return Module._load(path, this);
 };
 
@@ -434,7 +439,12 @@ Module.prototype._compile = function(content, filename) {
   var compiledWrapper = runInThisContext(wrapper, filename, true);
   if (global.v8debug) {
     if (!resolvedArgv) {
-      resolvedArgv = Module._resolveFilename(process.argv[1], null);
+      // we enter the repl if we're not given a filename argument.
+      if (process.argv[1]) {
+        resolvedArgv = Module._resolveFilename(process.argv[1], null);
+      } else {
+        resolvedArgv = 'repl';
+      }
     }
 
     // Set breakpoint on module start
@@ -468,32 +478,45 @@ Module._extensions['.js'] = function(module, filename) {
 // Native extension for .json
 Module._extensions['.json'] = function(module, filename) {
   var content = NativeModule.require('fs').readFileSync(filename, 'utf8');
-  module.exports = JSON.parse(stripBOM(content));
+  try {
+    module.exports = JSON.parse(stripBOM(content));
+  } catch (err) {
+    err.message = filename + ': ' + err.message;
+    throw err;
+  }
 };
 
 
 //Native extension for .node
-Module._extensions['.node'] = function(module, filename) {
-  process.dlopen(filename, module.exports);
-};
+Module._extensions['.node'] = process.dlopen;
 
 
 // bootstrap main module.
 Module.runMain = function() {
   // Load the main module--the command line argument.
   Module._load(process.argv[1], null, true);
+  // Handle any nextTicks added in the first tick of the program
+  process._tickCallback();
 };
 
 Module._initPaths = function() {
+  var isWindows = process.platform === 'win32';
+
+  if (isWindows) {
+    var homeDir = process.env.USERPROFILE;
+  } else {
+    var homeDir = process.env.HOME;
+  }
+
   var paths = [path.resolve(process.execPath, '..', '..', 'lib', 'node')];
 
-  if (process.env['HOME']) {
-    paths.unshift(path.resolve(process.env['HOME'], '.node_libraries'));
-    paths.unshift(path.resolve(process.env['HOME'], '.node_modules'));
+  if (homeDir) {
+    paths.unshift(path.resolve(homeDir, '.node_libraries'));
+    paths.unshift(path.resolve(homeDir, '.node_modules'));
   }
 
   if (process.env['NODE_PATH']) {
-    var splitter = process.platform === 'win32' ? ';' : ':';
+    var splitter = isWindows ? ';' : ':';
     paths = process.env['NODE_PATH'].split(splitter).concat(paths);
   }
 
