@@ -24,8 +24,7 @@ var util = require('util');
 var net = require('net');
 var url = require('url');
 var events = require('events');
-var Stream = require('stream');
-var END_OF_FILE = 42;
+var stream = require('stream');
 var assert = require('assert').ok;
 var constants = require('constants');
 
@@ -39,6 +38,19 @@ var DEFAULT_CIPHERS = 'ECDHE-RSA-AES128-SHA256:AES128-GCM-SHA256:' + // TLS 1.2
 exports.CLIENT_RENEG_LIMIT = 3;
 exports.CLIENT_RENEG_WINDOW = 600;
 
+exports.SLAB_BUFFER_SIZE = 10 * 1024 * 1024;
+
+exports.getCiphers = function() {
+  var names = process.binding('crypto').getSSLCiphers();
+  // Drop all-caps names in favor of their lowercase aliases,
+  var ctx = {};
+  names.forEach(function(name) {
+    if (/^[0-9A-Z\-]+$/.test(name)) name = name.toLowerCase();
+    ctx[name] = true;
+  });
+  return Object.getOwnPropertyNames(ctx).sort();
+};
+
 
 var debug;
 if (process.env.NODE_DEBUG && /tls/.test(process.env.NODE_DEBUG)) {
@@ -91,7 +103,14 @@ function checkServerIdentity(host, cert) {
     // The same applies to hostname with more than one wildcard,
     // if hostname has wildcard when wildcards are not allowed,
     // or if there are less than two dots after wildcard (i.e. *.com or *d.com)
-    if (/\*.*\*/.test(host) || !wildcards && /\*/.test(host) ||
+    //
+    // also
+    //
+    // "The client SHOULD NOT attempt to match a presented identifier in
+    // which the wildcard character comprises a label other than the
+    // left-most label (e.g., do not match bar.*.example.net)."
+    // RFC6125
+    if (!wildcards && /\*/.test(host) || /[\.\*].*\*/.test(host) ||
         /\*/.test(host) && !/\*.*\..+\..+/.test(host)) {
       return /$./;
     }
@@ -112,6 +131,7 @@ function checkServerIdentity(host, cert) {
   var dnsNames = [],
       uriNames = [],
       ips = [],
+      matchCN = true,
       valid = false;
 
   // There're several names to perform check against:
@@ -155,15 +175,25 @@ function checkServerIdentity(host, cert) {
 
     dnsNames = dnsNames.concat(uriNames);
 
-    // And only after check if hostname matches CN
-    // (because CN is deprecated, but should be used for compatiblity anyway)
-    var commonNames = cert.subject.CN;
-    if (Array.isArray(commonNames)) {
-      for (var i = 0, k = commonNames.length; i < k; ++i) {
-        dnsNames.push(regexpify(commonNames[i], false));
+    if (dnsNames.length > 0) matchCN = false;
+
+    // Match against Common Name (CN) only if no supported identifiers are
+    // present.
+    //
+    // "As noted, a client MUST NOT seek a match for a reference identifier
+    //  of CN-ID if the presented identifiers include a DNS-ID, SRV-ID,
+    //  URI-ID, or any application-specific identifier types supported by the
+    //  client."
+    // RFC6125
+    if (matchCN) {
+      var commonNames = cert.subject.CN;
+      if (Array.isArray(commonNames)) {
+        for (var i = 0, k = commonNames.length; i < k; ++i) {
+          dnsNames.push(regexpify(commonNames[i], true));
+        }
+      } else {
+        dnsNames.push(regexpify(commonNames, true));
       }
-    } else {
-      dnsNames.push(regexpify(commonNames, false));
     }
 
     valid = dnsNames.some(function(re) {
@@ -176,7 +206,6 @@ function checkServerIdentity(host, cert) {
 exports.checkServerIdentity = checkServerIdentity;
 
 
-
 function SlabBuffer() {
   this.create();
 }
@@ -184,20 +213,23 @@ function SlabBuffer() {
 
 SlabBuffer.prototype.create = function create() {
   this.isFull = false;
-  this.pool = new Buffer(10 * 1024 * 1024);
+  this.pool = new Buffer(exports.SLAB_BUFFER_SIZE);
   this.offset = 0;
   this.remaining = this.pool.length;
 };
 
 
-SlabBuffer.prototype.use = function use(context, fn) {
+SlabBuffer.prototype.use = function use(context, fn, size) {
   if (this.remaining === 0) {
     this.isFull = true;
     return 0;
   }
 
-  var bytes = fn.call(context, this.pool, this.offset, this.remaining);
+  var actualSize = this.remaining;
 
+  if (size !== null) actualSize = Math.min(size, actualSize);
+
+  var bytes = fn.call(context, this.pool, this.offset, actualSize);
   if (bytes > 0) {
     this.offset += bytes;
     this.remaining -= bytes;
@@ -209,96 +241,293 @@ SlabBuffer.prototype.use = function use(context, fn) {
 };
 
 
-var slabBuffer = new SlabBuffer();
+var slabBuffer = null;
 
 
 // Base class of both CleartextStream and EncryptedStream
-function CryptoStream(pair) {
-  Stream.call(this);
+function CryptoStream(pair, options) {
+  stream.Duplex.call(this, options);
 
   this.pair = pair;
+  this._pending = null;
+  this._pendingEncoding = '';
+  this._pendingCallback = null;
+  this._doneFlag = false;
+  this._retryAfterPartial = false;
+  this._halfRead = false;
+  this._sslOutCb = null;
+  this._resumingSession = false;
+  this._reading = true;
+  this._destroyed = false;
+  this._ended = false;
+  this._finished = false;
+  this._opposite = null;
 
-  this.readable = this.writable = true;
-
-  this._paused = false;
-  this._needDrain = false;
-  this._pending = [];
-  this._pendingCallbacks = [];
-  this._pendingBytes = 0;
+  if (slabBuffer === null) slabBuffer = new SlabBuffer();
   this._buffer = slabBuffer;
+
+  this.once('finish', onCryptoStreamFinish);
+
+  // net.Socket calls .onend too
+  this.once('end', onCryptoStreamEnd);
 }
-util.inherits(CryptoStream, Stream);
+util.inherits(CryptoStream, stream.Duplex);
 
 
-CryptoStream.prototype.write = function(data /* , encoding, cb */) {
-  if (this == this.pair.cleartext) {
-    debug('cleartext.write called with ' + data.length + ' bytes');
-  } else {
-    debug('encrypted.write called with ' + data.length + ' bytes');
-  }
+function onCryptoStreamFinish() {
+  this._finished = true;
 
-  if (!this.writable) {
-    throw new Error('CryptoStream is not writable');
-  }
+  if (this === this.pair.cleartext) {
+    debug('cleartext.onfinish');
+    if (this.pair.ssl) {
+      // Generate close notify
+      // NOTE: first call checks if client has sent us shutdown,
+      // second call enqueues shutdown into the BIO.
+      if (this.pair.ssl.shutdown() !== 1) {
+        if (this.pair.ssl && this.pair.ssl.error)
+          return this.pair.error();
 
-  var encoding, cb;
-
-  // parse arguments
-  if (typeof arguments[1] == 'string') {
-    encoding = arguments[1];
-    cb = arguments[2];
-  } else {
-    cb = arguments[1];
-  }
-
-
-  // Transform strings into buffers.
-  if (typeof data == 'string') {
-    data = new Buffer(data, encoding);
-  }
-
-  debug((this === this.pair.cleartext ? 'clear' : 'encrypted') + 'In data');
-
-  this._pending.push(data);
-  this._pendingCallbacks.push(cb);
-  this._pendingBytes += data.length;
-
-  this.pair._writeCalled = true;
-  this.pair.cycle();
-
-  // In the following cases, write() should return a false,
-  // then this stream should eventually emit 'drain' event.
-  //
-  // 1. There are pending data more than 128k bytes.
-  // 2. A forward stream shown below is paused.
-  //    A) EncryptedStream for CleartextStream.write().
-  //    B) CleartextStream for EncryptedStream.write().
-  //
-  if (!this._needDrain) {
-    if (this._pendingBytes >= 128 * 1024) {
-      this._needDrain = true;
-    } else {
-      if (this === this.pair.cleartext) {
-        this._needDrain = this.pair.encrypted._paused;
-      } else {
-        this._needDrain = this.pair.cleartext._paused;
+        this.pair.ssl.shutdown();
       }
+
+      if (this.pair.ssl && this.pair.ssl.error)
+        return this.pair.error();
+    }
+  } else {
+    debug('encrypted.onfinish');
+  }
+
+  // Try to read just to get sure that we won't miss EOF
+  if (this._opposite.readable) this._opposite.read(0);
+
+  if (this._opposite._ended) {
+    this._done();
+
+    // No half-close, sorry
+    if (this === this.pair.cleartext) this._opposite._done();
+  }
+}
+
+
+function onCryptoStreamEnd() {
+  this._ended = true;
+  if (this === this.pair.cleartext) {
+    debug('cleartext.onend');
+  } else {
+    debug('encrypted.onend');
+  }
+
+  if (this.onend) this.onend();
+}
+
+
+// NOTE: Called once `this._opposite` is set.
+CryptoStream.prototype.init = function init() {
+  var self = this;
+  this._opposite.on('sslOutEnd', function() {
+    if (self._sslOutCb) {
+      var cb = self._sslOutCb;
+      self._sslOutCb = null;
+      cb(null);
+    }
+  });
+};
+
+
+CryptoStream.prototype._write = function write(data, encoding, cb) {
+  assert(this._pending === null);
+
+  // Black-hole data
+  if (!this.pair.ssl) return cb(null);
+
+  // When resuming session don't accept any new data.
+  // And do not put too much data into openssl, before writing it from encrypted
+  // side.
+  //
+  // TODO(indutny): Remove magic number, use watermark based limits
+  if (!this._resumingSession &&
+      this._opposite._internallyPendingBytes() < 128 * 1024) {
+    // Write current buffer now
+    var written;
+    if (this === this.pair.cleartext) {
+      debug('cleartext.write called with ' + data.length + ' bytes');
+      written = this.pair.ssl.clearIn(data, 0, data.length);
+    } else {
+      debug('encrypted.write called with ' + data.length + ' bytes');
+      written = this.pair.ssl.encIn(data, 0, data.length);
+    }
+
+    // Handle and report errors
+    if (this.pair.ssl && this.pair.ssl.error) {
+      return cb(this.pair.error(true));
+    }
+
+    // Force SSL_read call to cycle some states/data inside OpenSSL
+    this.pair.cleartext.read(0);
+
+    // Cycle encrypted data
+    if (this.pair.encrypted._internallyPendingBytes())
+      this.pair.encrypted.read(0);
+
+    // Get NPN and Server name when ready
+    this.pair.maybeInitFinished();
+
+    // Whole buffer was written
+    if (written === data.length) {
+      if (this === this.pair.cleartext) {
+        debug('cleartext.write succeed with ' + written + ' bytes');
+      } else {
+        debug('encrypted.write succeed with ' + written + ' bytes');
+      }
+
+      // Invoke callback only when all data read from opposite stream
+      if (this._opposite._halfRead) {
+        assert(this._sslOutCb === null);
+        this._sslOutCb = cb;
+      } else {
+        cb(null);
+      }
+      return;
+    } else if (written !== 0 && written !== -1) {
+      assert(!this._retryAfterPartial);
+      this._retryAfterPartial = true;
+      this._write(data.slice(written), encoding, cb);
+      this._retryAfterPartial = false;
+      return;
+    }
+  } else {
+    debug('cleartext.write queue is full');
+
+    // Force SSL_read call to cycle some states/data inside OpenSSL
+    this.pair.cleartext.read(0);
+  }
+
+  // No write has happened
+  this._pending = data;
+  this._pendingEncoding = encoding;
+  this._pendingCallback = cb;
+
+  if (this === this.pair.cleartext) {
+    debug('cleartext.write queued with ' + data.length + ' bytes');
+  } else {
+    debug('encrypted.write queued with ' + data.length + ' bytes');
+  }
+};
+
+
+CryptoStream.prototype._writePending = function writePending() {
+  var data = this._pending,
+      encoding = this._pendingEncoding,
+      cb = this._pendingCallback;
+
+  this._pending = null;
+  this._pendingEncoding = '';
+  this._pendingCallback = null;
+  this._write(data, encoding, cb);
+};
+
+
+CryptoStream.prototype._read = function read(size) {
+  // XXX: EOF?!
+  if (!this.pair.ssl) return this.push(null);
+
+  // Wait for session to be resumed
+  // Mark that we're done reading, but don't provide data or EOF
+  if (this._resumingSession || !this._reading) return this.push('');
+
+  var out;
+  if (this === this.pair.cleartext) {
+    debug('cleartext.read called with ' + size + ' bytes');
+    out = this.pair.ssl.clearOut;
+  } else {
+    debug('encrypted.read called with ' + size + ' bytes');
+    out = this.pair.ssl.encOut;
+  }
+
+  var bytesRead = 0,
+      start = this._buffer.offset;
+  do {
+    var read = this._buffer.use(this.pair.ssl, out, size);
+    if (read > 0) {
+      bytesRead += read;
+      size -= read;
+    }
+
+    // Handle and report errors
+    if (this.pair.ssl && this.pair.ssl.error) {
+      this.pair.error();
+      break;
+    }
+
+    // Get NPN and Server name when ready
+    this.pair.maybeInitFinished();
+  } while (read > 0 && !this._buffer.isFull && bytesRead < size);
+
+  // Create new buffer if previous was filled up
+  var pool = this._buffer.pool;
+  if (this._buffer.isFull) this._buffer.create();
+
+  assert(bytesRead >= 0);
+
+  if (this === this.pair.cleartext) {
+    debug('cleartext.read succeed with ' + bytesRead + ' bytes');
+  } else {
+    debug('encrypted.read succeed with ' + bytesRead + ' bytes');
+  }
+
+  // Try writing pending data
+  if (this._pending !== null) this._writePending();
+  if (this._opposite._pending !== null) this._opposite._writePending();
+
+  if (bytesRead === 0) {
+    // EOF when cleartext has finished and we have nothing to read
+    if (this._opposite._finished && this._internallyPendingBytes() === 0) {
+      // Perform graceful shutdown
+      this._done();
+
+      // No half-open, sorry!
+      if (this === this.pair.cleartext)
+        this._opposite._done();
+
+      // EOF
+      this.push(null);
+    } else {
+      // Bail out
+      this.push('');
+    }
+  } else {
+    // Give them requested data
+    if (this.ondata) {
+      var self = this;
+      this.ondata(pool, start, start + bytesRead);
+
+      // Consume data automatically
+      // simple/test-https-drain fails without it
+      process.nextTick(function() {
+        self.read(bytesRead);
+      });
+    }
+    this.push(pool.slice(start, start + bytesRead));
+  }
+
+  // Let users know that we've some internal data to read
+  var halfRead = this._internallyPendingBytes() !== 0;
+
+  // Smart check to avoid invoking 'sslOutEnd' in the most of the cases
+  if (this._halfRead !== halfRead) {
+    this._halfRead = halfRead;
+
+    // Notify listeners about internal data end
+    if (!halfRead) {
+      if (this === this.pair.cleartext) {
+        debug('cleartext.sslOutEnd');
+      } else {
+        debug('encrypted.sslOutEnd');
+      }
+
+      this.emit('sslOutEnd');
     }
   }
-  return !this._needDrain;
-};
-
-
-CryptoStream.prototype.pause = function() {
-  debug('paused ' + (this == this.pair.cleartext ? 'cleartext' : 'encrypted'));
-  this._paused = true;
-};
-
-
-CryptoStream.prototype.resume = function() {
-  debug('resume ' + (this == this.pair.cleartext ? 'cleartext' : 'encrypted'));
-  this._paused = false;
-  this.pair.cycle();
 };
 
 
@@ -316,11 +545,9 @@ CryptoStream.prototype.setKeepAlive = function(enable, initialDelay) {
   if (this.socket) this.socket.setKeepAlive(enable, initialDelay);
 };
 
-
-CryptoStream.prototype.setEncoding = function(encoding) {
-  var StringDecoder = require('string_decoder').StringDecoder; // lazy load
-  this._decoder = new StringDecoder(encoding);
-};
+CryptoStream.prototype.__defineGetter__('bytesWritten', function() {
+  return this.socket ? this.socket.bytesWritten : 0;
+});
 
 
 // Example:
@@ -386,53 +613,87 @@ CryptoStream.prototype.getCipher = function(err) {
 };
 
 
-CryptoStream.prototype.end = function(d) {
-  if (this.pair._doneFlag) return;
-  if (!this.writable) return;
-
-  if (d) {
-    this.write(d);
+CryptoStream.prototype.end = function(chunk, encoding) {
+  if (this === this.pair.cleartext) {
+    debug('cleartext.end');
+  } else {
+    debug('encrypted.end');
   }
 
-  this._pending.push(END_OF_FILE);
-  this._pendingCallbacks.push(null);
-
-  // If this is an encrypted stream then we need to disable further 'data'
-  // events.
+  // Write pending data first
+  if (this._pending !== null) this._writePending();
 
   this.writable = false;
 
-  this.pair.cycle();
+  stream.Duplex.prototype.end.call(this, chunk, encoding);
 };
 
 
 CryptoStream.prototype.destroySoon = function(err) {
-  if (this.writable) {
-    this.end();
+  if (this === this.pair.cleartext) {
+    debug('cleartext.destroySoon');
   } else {
+    debug('encrypted.destroySoon');
+  }
+
+  if (this.writable)
+    this.end();
+
+  if (this._writableState.finished && this._opposite._ended) {
     this.destroy();
+  } else {
+    // Wait for both `finish` and `end` events to ensure that all data that
+    // was written on this side was read from the other side.
+    var self = this;
+    var waiting = 1;
+    this._opposite.once('end', finish);
+    if (!this._finished) {
+      this.once('finish', finish);
+      ++waiting;
+    }
+  }
+
+  function finish() {
+    if (--waiting === 0) self.destroy();
   }
 };
 
 
 CryptoStream.prototype.destroy = function(err) {
-  if (this.pair._doneFlag) return;
-  this.pair.destroy();
+  if (this._destroyed) return;
+  this._destroyed = true;
+  this.readable = this.writable = false;
+
+  // Destroy both ends
+  if (this === this.pair.cleartext) {
+    debug('cleartext.destroy');
+  } else {
+    debug('encrypted.destroy');
+  }
+  this._opposite.destroy();
+
+  var self = this;
+  process.nextTick(function() {
+    // Force EOF
+    self.push(null);
+
+    // Emit 'close' event
+    self.emit('close', err ? true : false);
+  });
 };
 
 
 CryptoStream.prototype._done = function() {
   this._doneFlag = true;
 
+  if (this === this.pair.encrypted && !this.pair._secureEstablished)
+    return this.pair.error();
+
   if (this.pair.cleartext._doneFlag &&
       this.pair.encrypted._doneFlag &&
       !this.pair._doneFlag) {
     // If both streams are done:
-    if (!this.pair._secureEstablished) {
-      this.pair.error();
-    } else {
-      this.pair.destroy();
-    }
+    this.pair.destroy();
   }
 };
 
@@ -455,182 +716,23 @@ Object.defineProperty(CryptoStream.prototype, 'readyState', {
 });
 
 
-// Move decrypted, clear data out into the application.
-// From the user's perspective this occurs as a 'data' event
-// on the pair.cleartext.
-// also
-// Move encrypted data to the stream. From the user's perspective this
-// occurs as a 'data' event on the pair.encrypted. Usually the application
-// will have some code which pipes the stream to a socket:
-//
-//   pair.encrypted.on('data', function (d) {
-//     socket.write(d);
-//   });
-//
-CryptoStream.prototype._push = function() {
-  if (this == this.pair.encrypted && !this.writable) {
-    // If the encrypted side got EOF, we do not attempt
-    // to write out data anymore.
-    return;
-  }
+function CleartextStream(pair, options) {
+  CryptoStream.call(this, pair, options);
 
-  while (!this._paused) {
-    var chunkBytes = 0,
-        bytesRead = 0,
-        start = this._buffer.offset;
-
-    do {
-      chunkBytes = this._buffer.use(this, this._pusher);
-      if (chunkBytes > 0) bytesRead += chunkBytes;
-
-      if (this.pair.ssl && this.pair.ssl.error) {
-        this.pair.error();
-        return;
-      }
-
-      this.pair.maybeInitFinished();
-
-    } while (chunkBytes > 0 && !this._buffer.isFull);
-
-    var pool = this._buffer.pool;
-
-    // Create new buffer if previous was filled up
-    if (this._buffer.isFull) this._buffer.create();
-
-    assert(bytesRead >= 0);
-
-    // Bail out if we didn't read any data.
-    if (bytesRead == 0) {
-      if (this._internallyPendingBytes() == 0 && this._destroyAfterPush) {
-        this._done();
-      }
-      return;
+  // This is a fake kludge to support how the http impl sits
+  // on top of net Sockets
+  var self = this;
+  this._handle = {
+    readStop: function() {
+      self._reading = false;
+    },
+    readStart: function() {
+      if (self._reading && self._readableState.length > 0) return;
+      self._reading = true;
+      self.read(0);
+      if (self._opposite.readable) self._opposite.read(0);
     }
-
-    var chunk = pool.slice(start, start + bytesRead);
-
-    if (this === this.pair.cleartext) {
-      debug('cleartext emit "data" with ' + bytesRead + ' bytes');
-    } else {
-      debug('encrypted emit "data" with ' + bytesRead + ' bytes');
-    }
-
-    if (this._decoder) {
-      var string = this._decoder.write(chunk);
-      if (string.length) this.emit('data', string);
-    } else {
-      this.emit('data', chunk);
-    }
-
-    // Optimization: emit the original buffer with end points
-    if (this.ondata) this.ondata(pool, start, start + bytesRead);
-  }
-};
-
-
-// Push in any clear data coming from the application.
-// This arrives via some code like this:
-//
-//   pair.cleartext.write("hello world");
-//
-// also
-//
-// Push in incoming encrypted data from the socket.
-// This arrives via some code like this:
-//
-//   socket.on('data', function (d) {
-//     pair.encrypted.write(d)
-//   });
-//
-CryptoStream.prototype._pull = function() {
-  var havePending = this._pending.length > 0;
-
-  assert(havePending || this._pendingBytes == 0);
-
-  while (this._pending.length > 0) {
-    if (!this.pair.ssl) break;
-
-    var tmp = this._pending.shift();
-    var cb = this._pendingCallbacks.shift();
-
-    assert(this._pending.length === this._pendingCallbacks.length);
-
-    if (tmp === END_OF_FILE) {
-      // Sending EOF
-      if (this === this.pair.encrypted) {
-        debug('end encrypted ' + this.pair.fd);
-        this.pair.cleartext._destroyAfterPush = true;
-      } else {
-        // CleartextStream
-        assert(this === this.pair.cleartext);
-        debug('end cleartext');
-
-        this.pair.ssl.shutdown();
-
-        // TODO check if we get EAGAIN From shutdown, would have to do it
-        // again. should unshift END_OF_FILE back onto pending and wait for
-        // next cycle.
-
-        this.pair.encrypted._destroyAfterPush = true;
-      }
-      this.pair.cycle();
-      this._done();
-      return;
-    }
-
-    if (tmp.length == 0) continue;
-
-    var rv = this._puller(tmp);
-
-    if (this.pair.ssl && this.pair.ssl.error) {
-      this.pair.error();
-      return;
-    }
-
-    this.pair.maybeInitFinished();
-
-    if (rv === 0 || rv < 0) {
-      this._pending.unshift(tmp);
-      this._pendingCallbacks.unshift(cb);
-      break;
-    }
-
-    this._pendingBytes -= tmp.length;
-    assert(this._pendingBytes >= 0);
-
-    if (cb) cb();
-
-    assert(rv === tmp.length);
-  }
-
-  // If pending data has cleared, 'drain' event should be emitted
-  // after write() returns a false.
-  // Except when a forward stream shown below is paused.
-  //   A) EncryptedStream for CleartextStream._pull().
-  //   B) CleartextStream for EncryptedStream._pull().
-  //
-  if (this._needDrain && this._pending.length === 0) {
-    var paused;
-    if (this === this.pair.cleartext) {
-      paused = this.pair.encrypted._paused;
-    } else {
-      paused = this.pair.cleartext._paused;
-    }
-    if (!paused) {
-      debug('drain ' + (this === this.pair.cleartext ? 'clear' : 'encrypted'));
-      var self = this;
-      process.nextTick(function() {
-        self.emit('drain');
-      });
-      this._needDrain = false;
-      if (this.__destroyOnDrain) this.end();
-    }
-  }
-};
-
-
-function CleartextStream(pair) {
-  CryptoStream.call(this, pair);
+  };
 }
 util.inherits(CleartextStream, CryptoStream);
 
@@ -644,22 +746,11 @@ CleartextStream.prototype._internallyPendingBytes = function() {
 };
 
 
-CleartextStream.prototype._puller = function(b) {
-  debug('clearIn ' + b.length + ' bytes');
-  return this.pair.ssl.clearIn(b, 0, b.length);
-};
-
-
-CleartextStream.prototype._pusher = function(pool, offset, length) {
-  debug('reading from clearOut');
-  if (!this.pair.ssl) return -1;
-  return this.pair.ssl.clearOut(pool, offset, length);
-};
-
 CleartextStream.prototype.address = function() {
   return this.socket && this.socket.address();
 };
 
+
 CleartextStream.prototype.__defineGetter__('remoteAddress', function() {
   return this.socket && this.socket.remoteAddress;
 });
@@ -669,9 +760,8 @@ CleartextStream.prototype.__defineGetter__('remotePort', function() {
   return this.socket && this.socket.remotePort;
 });
 
-
-function EncryptedStream(pair) {
-  CryptoStream.call(this, pair);
+function EncryptedStream(pair, options) {
+  CryptoStream.call(this, pair, options);
 }
 util.inherits(EncryptedStream, CryptoStream);
 
@@ -685,35 +775,28 @@ EncryptedStream.prototype._internallyPendingBytes = function() {
 };
 
 
-EncryptedStream.prototype._puller = function(b) {
-  debug('writing from encIn');
-  return this.pair.ssl.encIn(b, 0, b.length);
-};
-
-
-EncryptedStream.prototype._pusher = function(pool, offset, length) {
-  debug('reading from encOut');
-  if (!this.pair.ssl) return -1;
-  return this.pair.ssl.encOut(pool, offset, length);
-};
-
-
 function onhandshakestart() {
   debug('onhandshakestart');
 
-  var self = this, ssl = this.ssl;
+  var self = this;
+  var ssl = self.ssl;
+  var now = Date.now();
 
-  if (ssl.timer === null) {
-    ssl.timer = setTimeout(function timeout() {
-      ssl.handshakes = 0;
-      ssl.timer = null;
-    }, exports.CLIENT_RENEG_WINDOW * 1000);
+  assert(now >= ssl.lastHandshakeTime);
+
+  if ((now - ssl.lastHandshakeTime) >= exports.CLIENT_RENEG_WINDOW * 1000) {
+    ssl.handshakes = 0;
   }
-  else if (++ssl.handshakes > exports.CLIENT_RENEG_LIMIT) {
+
+  var first = (ssl.lastHandshakeTime === 0);
+  ssl.lastHandshakeTime = now;
+  if (first) return;
+
+  if (++ssl.handshakes > exports.CLIENT_RENEG_LIMIT) {
     // Defer the error event to the next tick. We're being called from OpenSSL's
     // state machine and OpenSSL is not re-entrant. We cannot allow the user's
     // callback to destroy the connection right now, it would crash and burn.
-    process.nextTick(function() {
+    setImmediate(function() {
       var err = new Error('TLS session renegotiation attack detected.');
       if (self.cleartext) self.cleartext.emit('error', err);
     });
@@ -727,6 +810,39 @@ function onhandshakedone() {
 }
 
 
+function onclienthello(hello) {
+  var self = this,
+      once = false;
+
+  this._resumingSession = true;
+  function callback(err, session) {
+    if (once) return;
+    once = true;
+
+    if (err) return self.socket.destroy(err);
+
+    self.ssl.loadSession(session);
+
+    // Cycle data
+    self._resumingSession = false;
+    self.cleartext.read(0);
+    self.encrypted.read(0);
+  }
+
+  if (hello.sessionId.length <= 0 ||
+      !this.server ||
+      !this.server.emit('resumeSession', hello.sessionId, callback)) {
+    callback(null, null);
+  }
+}
+
+
+function onnewsession(key, session) {
+  if (!this.server) return;
+  this.server.emit('newSession', key, session);
+}
+
+
 /**
  * Provides a pair of streams to do encrypted communication.
  */
@@ -747,11 +863,13 @@ function SecurePair(credentials, isServer, requestCert, rejectUnauthorized,
 
   events.EventEmitter.call(this);
 
+  this.server = options.server;
   this._secureEstablished = false;
   this._isServer = isServer ? true : false;
   this._encWriteState = true;
   this._clearWriteState = true;
   this._doneFlag = false;
+  this._destroying = false;
 
   if (!credentials) {
     this.credentials = crypto.createCredentials();
@@ -769,15 +887,18 @@ function SecurePair(credentials, isServer, requestCert, rejectUnauthorized,
   this._requestCert = requestCert ? true : false;
 
   this.ssl = new Connection(this.credentials.context,
-      this._isServer ? true : false,
-      this._isServer ? this._requestCert : options.servername,
-      this._rejectUnauthorized);
+                            this._isServer ? true : false,
+                            this._isServer ? this._requestCert :
+                                             options.servername,
+                            this._rejectUnauthorized);
 
   if (this._isServer) {
     this.ssl.onhandshakestart = onhandshakestart.bind(this);
     this.ssl.onhandshakedone = onhandshakedone.bind(this);
+    this.ssl.onclienthello = onclienthello.bind(this);
+    this.ssl.onnewsession = onnewsession.bind(this);
+    this.ssl.lastHandshakeTime = 0;
     this.ssl.handshakes = 0;
-    this.ssl.timer = null;
   }
 
   if (process.features.tls_sni) {
@@ -793,17 +914,22 @@ function SecurePair(credentials, isServer, requestCert, rejectUnauthorized,
   }
 
   /* Acts as a r/w stream to the cleartext side of the stream. */
-  this.cleartext = new CleartextStream(this);
+  this.cleartext = new CleartextStream(this, options.cleartext);
 
   /* Acts as a r/w stream to the encrypted side of the stream. */
-  this.encrypted = new EncryptedStream(this);
+  this.encrypted = new EncryptedStream(this, options.encrypted);
+
+  /* Let streams know about each other */
+  this.cleartext._opposite = this.encrypted;
+  this.encrypted._opposite = this.cleartext;
+  this.cleartext.init();
+  this.encrypted.init();
 
   process.nextTick(function() {
     /* The Connection may be destroyed by an abort call */
     if (self.ssl) {
       self.ssl.start();
     }
-    self.cycle();
   });
 }
 
@@ -822,81 +948,6 @@ exports.createSecurePair = function(credentials,
 };
 
 
-
-
-/* Attempt to cycle OpenSSLs buffers in various directions.
- *
- * An SSL Connection can be viewed as four separate piplines,
- * interacting with one has no connection to the behavoir of
- * any of the other 3 -- This might not sound reasonable,
- * but consider things like mid-stream renegotiation of
- * the ciphers.
- *
- * The four pipelines, using terminology of the client (server is just
- * reversed):
- *  (1) Encrypted Output stream (Writing encrypted data to peer)
- *  (2) Encrypted Input stream (Reading encrypted data from peer)
- *  (3) Cleartext Output stream (Decrypted content from the peer)
- *  (4) Cleartext Input stream (Cleartext content to send to the peer)
- *
- * This function attempts to pull any available data out of the Cleartext
- * input stream (4), and the Encrypted input stream (2).  Then it pushes any
- * data available from the cleartext output stream (3), and finally from the
- * Encrypted output stream (1)
- *
- * It is called whenever we do something with OpenSSL -- post reciving
- * content, trying to flush, trying to change ciphers, or shutting down the
- * connection.
- *
- * Because it is also called everywhere, we also check if the connection has
- * completed negotiation and emit 'secure' from here if it has.
- */
-SecurePair.prototype.cycle = function(depth) {
-  if (this._doneFlag) return;
-
-  depth = depth ? depth : 0;
-
-  if (depth == 0) this._writeCalled = false;
-
-  var established = this._secureEstablished;
-
-  if (!this.cycleEncryptedPullLock) {
-    this.cycleEncryptedPullLock = true;
-    debug('encrypted._pull');
-    this.encrypted._pull();
-    this.cycleEncryptedPullLock = false;
-  }
-
-  if (!this.cycleCleartextPullLock) {
-    this.cycleCleartextPullLock = true;
-    debug('cleartext._pull');
-    this.cleartext._pull();
-    this.cycleCleartextPullLock = false;
-  }
-
-  if (!this.cycleCleartextPushLock) {
-    this.cycleCleartextPushLock = true;
-    debug('cleartext._push');
-    this.cleartext._push();
-    this.cycleCleartextPushLock = false;
-  }
-
-  if (!this.cycleEncryptedPushLock) {
-    this.cycleEncryptedPushLock = true;
-    debug('encrypted._push');
-    this.encrypted._push();
-    this.cycleEncryptedPushLock = false;
-  }
-
-  if ((!established && this._secureEstablished) ||
-      (depth == 0 && this._writeCalled)) {
-    // If we were not established but now we are, let's cycle again.
-    // Or if there is some data to write...
-    this.cycle(depth + 1);
-  }
-};
-
-
 SecurePair.prototype.maybeInitFinished = function() {
   if (this.ssl && !this._secureEstablished && this.ssl.isInitFinished()) {
     if (process.features.tls_npn) {
@@ -915,54 +966,48 @@ SecurePair.prototype.maybeInitFinished = function() {
 
 
 SecurePair.prototype.destroy = function() {
-  var self = this;
+  if (this._destroying) return;
 
   if (!this._doneFlag) {
+    debug('SecurePair.destroy');
+    this._destroying = true;
+
+    // SecurePair should be destroyed only after it's streams
+    this.cleartext.destroy();
+    this.encrypted.destroy();
+
     this._doneFlag = true;
-
-    if (this.ssl.timer) {
-      clearTimeout(this.ssl.timer);
-      this.ssl.timer = null;
-    }
-
     this.ssl.error = null;
     this.ssl.close();
     this.ssl = null;
-
-    self.encrypted.writable = self.encrypted.readable = false;
-    self.cleartext.writable = self.cleartext.readable = false;
-
-    process.nextTick(function() {
-      self.cleartext.emit('end');
-      self.encrypted.emit('close');
-      self.cleartext.emit('close');
-    });
   }
 };
 
 
-SecurePair.prototype.error = function() {
+SecurePair.prototype.error = function(returnOnly) {
+  var err = this.ssl.error;
+  this.ssl.error = null;
+
   if (!this._secureEstablished) {
-    var error = this.ssl.error;
-    if (!error) {
-      error = new Error('socket hang up');
-      error.code = 'ECONNRESET';
+    // Emit ECONNRESET instead of zero return
+    if (!err || err.message === 'ZERO_RETURN') {
+      var connReset = new Error('socket hang up');
+      connReset.code = 'ECONNRESET';
+      connReset.sslError = err && err.message;
+
+      err = connReset;
     }
     this.destroy();
-    this.emit('error', error);
+    if (!returnOnly) this.emit('error', err);
+  } else if (this._isServer &&
+             this._rejectUnauthorized &&
+             /peer did not return a certificate/.test(err.message)) {
+    // Not really an error.
+    this.destroy();
   } else {
-    var err = this.ssl.error;
-    this.ssl.error = null;
-
-    if (this._isServer &&
-        this._rejectUnauthorized &&
-        /peer did not return a certificate/.test(err.message)) {
-      // Not really an error.
-      this.destroy();
-    } else {
-      this.cleartext.emit('error', err);
-    }
+    if (!returnOnly) this.cleartext.emit('error', err);
   }
+  return err;
 };
 
 // TODO: support anonymous (nocert) and PSK
@@ -1004,13 +1049,13 @@ SecurePair.prototype.error = function() {
 //   If true clients whose certificates are invalid for any reason will not
 //   be allowed to make connections. If false, they will simply be marked as
 //   unauthorized but secure communication will continue. By default this is
-//   false.
+//   true.
 //
 //
 //
 // Options:
 // - requestCert. Send verify request. Default to false.
-// - rejectUnauthorized. Boolean, default to false.
+// - rejectUnauthorized. Boolean, default to true.
 // - key. string.
 // - cert: string.
 // - ca: string or array of strings.
@@ -1059,6 +1104,10 @@ function Server(/* [options], listener */) {
   // Handle option defaults:
   this.setOptions(options);
 
+  if (!self.pfx && (!self.cert || !self.key)) {
+    throw new Error('Missing PFX or certificate + private key.');
+  }
+
   var sharedCreds = crypto.createCredentials({
     pfx: self.pfx,
     key: self.key,
@@ -1072,6 +1121,12 @@ function Server(/* [options], listener */) {
     sessionIdContext: self.sessionIdContext
   });
 
+  var timeout = options.handshakeTimeout || (120 * 1000);
+
+  if (typeof timeout !== 'number') {
+    throw new TypeError('handshakeTimeout must be a number');
+  }
+
   // constructor call
   net.Server.call(this, function(socket) {
     var creds = crypto.createCredentials(null, sharedCreds.context);
@@ -1081,14 +1136,29 @@ function Server(/* [options], listener */) {
                               self.requestCert,
                               self.rejectUnauthorized,
                               {
+                                server: self,
                                 NPNProtocols: self.NPNProtocols,
-                                SNICallback: self.SNICallback
+                                SNICallback: self.SNICallback,
+
+                                // Stream options
+                                cleartext: self._cleartext,
+                                encrypted: self._encrypted
                               });
 
     var cleartext = pipe(pair, socket);
     cleartext._controlReleased = false;
 
-    pair.on('secure', function() {
+    function listener() {
+      pair.emit('error', new Error('TLS handshake timeout'));
+    }
+
+    if (timeout > 0) {
+      socket.setTimeout(timeout, listener);
+    }
+
+    pair.once('secure', function() {
+      socket.setTimeout(0, listener);
+
       pair.cleartext.authorized = false;
       pair.cleartext.npnProtocol = pair.npnProtocol;
       pair.cleartext.servername = pair.servername;
@@ -1116,7 +1186,7 @@ function Server(/* [options], listener */) {
       }
     });
     pair.on('error', function(err) {
-      self.emit('clientError', err);
+      self.emit('clientError', err, this);
     });
   });
 
@@ -1171,6 +1241,8 @@ Server.prototype.setOptions = function(options) {
                                   .update(process.argv.join(' '))
                                   .digest('hex');
   }
+  if (options.cleartext) this.cleartext = options.cleartext;
+  if (options.encrypted) this.encrypted = options.encrypted;
 };
 
 // SNI Contexts High-Level API
@@ -1233,6 +1305,11 @@ exports.connect = function(/* [port, host], options, cb */) {
   var options = args[0];
   var cb = args[1];
 
+  var defaults = {
+    rejectUnauthorized: '0' !== process.env.NODE_TLS_REJECT_UNAUTHORIZED
+  };
+  options = util._extend(defaults, options || {});
+
   var socket = options.socket ? options.socket : new net.Stream();
 
   var sslcontext = crypto.createCredentials(options);
@@ -1243,11 +1320,16 @@ exports.connect = function(/* [port, host], options, cb */) {
                             options.rejectUnauthorized === true ? true : false,
                             {
                               NPNProtocols: this.NPNProtocols,
-                              servername: hostname
+                              servername: hostname,
+                              cleartext: options.cleartext,
+                              encrypted: options.encrypted
                             });
 
   if (options.session) {
-    pair.ssl.setSession(options.session);
+    var session = options.session;
+    if (typeof session === 'string')
+      session = new Buffer(session, 'binary');
+    pair.ssl.setSession(session);
   }
 
   var cleartext = pipe(pair, socket);
@@ -1307,12 +1389,37 @@ function pipe(pair, socket) {
   pair.encrypted.pipe(socket);
   socket.pipe(pair.encrypted);
 
+  pair.encrypted.on('close', function() {
+    process.nextTick(function() {
+      // Encrypted should be unpiped from socket to prevent possible
+      // write after destroy.
+      pair.encrypted.unpipe(socket);
+      socket.destroy();
+    });
+  });
+
   pair.fd = socket.fd;
   var cleartext = pair.cleartext;
   cleartext.socket = socket;
   cleartext.encrypted = pair.encrypted;
   cleartext.authorized = false;
 
+  // cycle the data whenever the socket drains, so that
+  // we can pull some more into it.  normally this would
+  // be handled by the fact that pipe() triggers read() calls
+  // on writable.drain, but CryptoStreams are a bit more
+  // complicated.  Since the encrypted side actually gets
+  // its data from the cleartext side, we have to give it a
+  // light kick to get in motion again.
+  socket.on('drain', function() {
+    if (pair.encrypted._pending)
+      pair.encrypted._writePending();
+    if (pair.cleartext._pending)
+      pair.cleartext._writePending();
+    pair.encrypted.read(0);
+    pair.cleartext.read(0);
+  });
+
   function onerror(e) {
     if (cleartext._controlReleased) {
       cleartext.emit('error', e);
@@ -1321,7 +1428,6 @@ function pipe(pair, socket) {
 
   function onclose() {
     socket.removeListener('error', onerror);
-    socket.removeListener('end', onclose);
     socket.removeListener('timeout', ontimeout);
   }
 
