@@ -22,10 +22,10 @@
 #ifndef node_script_h
 #define node_script_h
 
-#include <node.h>
-#include <node_object_wrap.h>
-#include <v8.h>
-#include <ev.h>
+#include "node.h"
+#include "node_object_wrap.h"
+#include "v8.h"
+#include "uv.h"
 
 namespace node {
 
